#include "parser.h"
#include "aliases.h"
#include "dstring.h"
#include "json_entities.h"
#include "lexer.h"
#include <stdio.h>
#include <stdlib.h>

INTERNAL jentity_t *current = NULL;

INTERNAL jcoll_t *get_collection_from_entity(const jentity_t *entity);
INTERNAL jentity_t *create_new_single_entity(const jval_t value,
                                             jentity_t *parent);
INTERNAL jentity_t *create_new_pair_entity(dstr_t *key, const jval_t value,
                                           jentity_t *parent);
INTERNAL void add_key(jentity_t **entity, dstr_t *key, const jval_t *value);
INTERNAL jentity_t *add_value(jentity_t **entity, const jval_t *value);
INTERNAL void add_collection(jentity_t **entity, const jval_t *value);

void parse_token(token_t token) {
  switch (token.type) {
  case TK_L_BRACE:
  case TK_L_BRACKET: {
    jval_t value = (jval_t){
        .type = JVAL_COLLECTION,
        .collection = (jcoll_t *)malloc(sizeof(jcoll_t)),
    };

    if (token.type == TK_L_BRACE) {
      value.collection->type = JCOLL_OBJECT;
    } else {
      value.collection->type = JCOLL_ARRAY;
    }

    value.collection->size = 0;
    value.collection->begin = NULL;
    value.collection->end = NULL;

    add_collection(&current, &value);

    break;
  }
  case TK_R_BRACE:
  case TK_R_BRACKET:
    if (current->parent) {
      current = current->parent;
    }

    break;
  case TK_STR_KEY: {
    jval_t value = (jval_t){.type = JVAL_EMPTY, .null_val = NULL};

    add_key(&current, dstr_from_string(token.value.string), &value);

    break;
  }
  case TK_NULL: {
    jval_t value = (jval_t){.type = JVAL_NULL, .null_val = NULL};

    add_value(&current, &value);

    break;
  }
  case TK_BOOL: {
    jval_t value =
        (jval_t){.type = JVAL_BOOLEAN, .boolean = token.value.boolean};

    add_value(&current, &value);

    break;
  }
  case TK_STR_VAL: {
    jval_t value = (jval_t){.type = JVAL_STRING,
                            .string = dstr_from_string(token.value.string)};

    add_value(&current, &value);

    break;
  }
  case TK_INTEGER: {
    jval_t value =
        (jval_t){.type = JVAL_INTEGER, .num_int = token.value.num_int};

    add_value(&current, &value);

    break;
  }
  case TK_DOUBLE: {
    jval_t value =
        (jval_t){.type = JVAL_DOUBLE, .num_dbl = token.value.num_frac};

    add_value(&current, &value);

    break;
  }
  case TK_NO_TOKEN:
    break;
  }
}

jcoll_t *get_collection_from_entity(const jentity_t *entity) {
  return entity->type == JENTITY_SINGLE ? entity->value.collection
                                        : entity->pair.value.collection;
}

jentity_t *create_new_single_entity(const jval_t value, jentity_t *parent) {
  jentity_t *entity = (jentity_t *)malloc(sizeof(jentity_t));

  if (!entity) {
    return NULL;
  }

  entity->type = JENTITY_SINGLE;
  entity->value = value;
  entity->parent = parent;
  entity->next = NULL;

  return entity;
}

jentity_t *create_new_pair_entity(dstr_t *key, const jval_t value,
                                  jentity_t *parent) {
  jentity_t *entity = (jentity_t *)malloc(sizeof(jentity_t));

  if (!entity) {
    return NULL;
  }

  entity->type = JENTITY_PAIR;
  entity->pair.key = key;
  entity->pair.value = value;
  entity->parent = parent;
  entity->next = NULL;

  return entity;
}

void add_key(jentity_t **entity, dstr_t *key, const jval_t *value) {
  jcoll_t *collection = get_collection_from_entity(*entity);

  if (!collection) {
    return;
  }

  if (!(collection->end)) {
    collection->begin = collection->end =
        create_new_pair_entity(key, *value, *entity);

    collection->size = 1;
  } else {
    jentity_t *new_entity = create_new_pair_entity(key, *value, *entity);

    collection->end->next = new_entity;

    collection->end = new_entity;

    ++(collection->size);
  }
}

jentity_t *add_value(jentity_t **entity, const jval_t *value) {
  jcoll_t *collection = get_collection_from_entity(*entity);

  if (!collection) {
    return NULL;
  }

  if (!(collection->end)) {
    collection->begin = collection->end =
        create_new_single_entity(*value, *entity);

    collection->size = 1;
  } else {
    if (collection->end->type == JENTITY_PAIR &&
        collection->end->pair.value.type == JVAL_EMPTY) {
      collection->end->pair.value = *value;
    } else {
      jentity_t *new_entity = create_new_single_entity(*value, *entity);

      collection->end->next = new_entity;

      collection->end = new_entity;

      ++(collection->size);
    }
  }

  return collection->end;
}

void add_collection(jentity_t **entity, const jval_t *value) {
  if (!(*entity)) {
    *entity = create_new_single_entity(*value, NULL);

    return;
  }

  *entity = add_value(entity, value);
}
